import importlib
import inspect
import typing
import pydantic as pydantic_

from dataclasses import dataclass

import typing_extensions
from annotated_types import GroupedMetadata, BaseMetadata
from pydantic_core import core_schema
from pydantic_core.core_schema import ValidationInfo, SerializerFunctionWrapHandler

from typenum.core import TypEnumMeta, _TypEnum, TypEnumContent

__all__ = [
    "Rename",
    "FieldMetadata",
    "TypEnumPydantic",
    "TypEnumPydanticMeta",
    "eval_content_type",
]

from typenum.pydantic.serialization import AdjacentlyTagged, InternallyTagged, ExternallyTagged
from typenum.pydantic.serialization.tagged import TaggedSerialization


@dataclass(frozen=True, slots=True)
class Rename(BaseMetadata):
    value: str


@dataclass
class FieldMetadata(GroupedMetadata):
    rename: typing.Optional[str] = None

    def __iter__(self) -> typing.Iterator[BaseMetadata]:
        if self.rename is not None:
            yield Rename(self.rename)


def eval_content_type(cls: type['TypEnumPydantic[TypEnumContent]']) -> type:
    # Eval annotation into real object
    base = cls.__orig_bases__[0]  # type: ignore
    module = importlib.import_module(base.__module__)
    return eval(cls.__content_type__, module.__dict__)  # type: ignore


class TypEnumPydanticMeta(TypEnumMeta):
    __serialization__: TaggedSerialization

    def __new__(
            cls,
            cls_name: str,
            bases: tuple[typing.Any],
            class_dict: dict[str, typing.Any],
            variant: typing.Optional[str] = None,
            content: typing.Optional[str] = None,
    ) -> typing.Any:
        enum_class = super().__new__(cls, cls_name, bases, class_dict)
        if enum_class.__annotations__.get("__abstract__"):
            return enum_class

        enum_class.__full_variant_name__ = cls_name
        enum_class.__variant_name__ = cls_name

        if enum_class.__is_variant__:
            return enum_class

        enum_class.__names_serialization__ = dict()
        enum_class.__names_deserialization__ = dict()

        if variant is not None and content is not None:
            enum_class.__serialization__ = AdjacentlyTagged(variant, content)
        elif variant is not None:
            enum_class.__serialization__ = InternallyTagged(variant)
        else:
            enum_class.__serialization__ = ExternallyTagged()

        annotation: typing.Union[type[typing_extensions.Annotated[typing.Any, BaseMetadata]], type]
        for attr, annotation in enum_class.__annotations__.items():
            if not hasattr(annotation, "__args__"):
                continue

            enum_variant = getattr(enum_class, attr)
            if isinstance(enum_variant.__content_type__, str):
                try:
                    enum_variant.__content_type__ = eval_content_type(enum_variant)
                except NameError:
                    ...

            if isinstance(annotation, typing._AnnotatedAlias):  # type: ignore
                metadata: list[typing.Union[BaseMetadata, GroupedMetadata]] = []
                for v in annotation.__metadata__:
                    if isinstance(v, FieldMetadata):
                        metadata.extend(v)
                    else:
                        metadata.append(v)

                for __meta__ in metadata:
                    if isinstance(__meta__, Rename):
                        if __meta__.value in enum_class.__names_deserialization__:
                            raise ValueError(f"{cls_name}: Two or many field renamed to `{__meta__.value}`")

                        enum_class.__names_serialization__[attr] = __meta__.value
                        enum_class.__names_deserialization__[__meta__.value] = attr

        return enum_class


class TypEnumPydantic(_TypEnum[TypEnumContent], metaclass=TypEnumPydanticMeta):
    __abstract__: typing_extensions.Never

    __names_serialization__: typing.ClassVar[dict[str, str]]
    __names_deserialization__: typing.ClassVar[dict[str, str]]

    __serialization__: typing.ClassVar[TaggedSerialization]

    @classmethod
    def __variant_constructor__(
            cls: type["TypEnumPydantic[TypEnumContent]"],
            value: typing.Any,
            info: ValidationInfo,
    ) -> "TypEnumPydantic[TypEnumContent]":
        # Resolve types when __content_type__ declare after cls declaration
        __content_type__ = cls.__content_type__
        if isinstance(__content_type__, str):
            __content_type__ = eval_content_type(cls)

        if inspect.isclass(__content_type__) and issubclass(__content_type__, TypEnumPydantic):
            value = cls.__python_value_restore__(value, info)

        return cls(value)

    @classmethod
    def __get_pydantic_core_schema__(
            cls: type["TypEnumPydantic[TypEnumContent]"],
            source_type: typing.Any,
            handler: pydantic_.GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return cls.__serialization__.__get_pydantic_core_schema__(cls, source_type, handler)

    @classmethod
    def __python_value_restore__(
            cls: type["TypEnumPydantic[TypEnumContent]"],
            input_value: typing.Any,
            info: ValidationInfo,
    ) -> typing.Any:
        return cls.__serialization__.__python_value_restore__(cls, input_value, info)

    @classmethod
    def __pydantic_serialization__(
            cls: type["TypEnumPydantic[TypEnumContent]"],
            model: typing.Any,
            serializer: SerializerFunctionWrapHandler,
    ) -> typing.Any:
        return cls.__serialization__.__pydantic_serialization__(cls, model, serializer)



from typenum.core import TypEnum, TypEnumContent, NoValue

__all__ = [
    "NoValue",
    "TypEnum",
    "TypEnumContent",
]

__package_name__ = "typenum"
__version__ = "0.3.0"
__description__ = "Type-containing enumeration"



import pydantic

if tuple(map(int, pydantic.version.VERSION.split('.'))) < (2, 9, 0):
    raise ValueError("Pydantic version must be >=2.9.0")


from .core import (
    Rename,
    TypEnumPydantic,
    FieldMetadata,
)


__all__ = [
    "FieldMetadata",
    "Rename",
    "TypEnumPydantic",
]



from .externally import ExternallyTagged
from .adjacently import AdjacentlyTagged
from .internally import InternallyTagged

__all__ = [
    "ExternallyTagged",
    "AdjacentlyTagged",
    "InternallyTagged",
]



import inspect
import typing

import pydantic as pydantic_
from pydantic_core import CoreSchema, core_schema
from pydantic_core.core_schema import SerializerFunctionWrapHandler, ValidationInfo

from typenum.core import TypEnumContent, NoValue
from typenum.pydantic.serialization.tagged import TaggedSerialization

if typing.TYPE_CHECKING:
    from ..core import TypEnumPydantic  # type: ignore

__all__ = [
    "ExternallyTagged",
]


class ExternallyTagged(TaggedSerialization):
    def __get_pydantic_core_schema__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            _source_type: typing.Any,
            handler: pydantic_.GetCoreSchemaHandler,
    ) -> CoreSchema:
        from typenum.pydantic.core import TypEnumPydantic

        json_schema_attrs = {}
        other_schemas = []
        for attr in kls.__variants__.values():
            enum_variant: type[TypEnumPydantic[TypEnumContent]] = getattr(kls, attr)
            attr = kls.__names_serialization__.get(attr, attr)

            is_typenum_variant = (
                    inspect.isclass(enum_variant.__content_type__) and
                    issubclass(enum_variant.__content_type__, TypEnumPydantic)
            )

            item_schema: core_schema.CoreSchema
            if is_typenum_variant or enum_variant.__content_type__ is NoValue:
                if enum_variant.__content_type__ is NoValue:
                    other_schemas.append(core_schema.str_schema(pattern=attr))
                    continue
                else:
                    kls_: type = enum_variant.__content_type__  # type: ignore
                    schema_definition = core_schema.definition_reference_schema(f"{kls_.__name__}:{id(kls_)}")
                    item_schema = core_schema.definitions_schema(
                        schema=schema_definition,
                        definitions=[
                            core_schema.any_schema(ref=f"{kls_.__name__}:{id(kls_)}")
                        ],
                    )

            else:
                item_schema = handler.generate_schema(enum_variant.__content_type__)

            json_schema_attrs[attr] = core_schema.typed_dict_field(item_schema, required=False)

        schemas = [core_schema.typed_dict_schema(json_schema_attrs), *other_schemas]

        return core_schema.json_or_python_schema(
            json_schema=core_schema.with_info_after_validator_function(
                kls.__python_value_restore__,
                core_schema.union_schema([*schemas]),
            ),
            python_schema=core_schema.with_info_after_validator_function(
                kls.__python_value_restore__,
                core_schema.union_schema([*schemas, core_schema.any_schema()]),
            ),
            serialization=core_schema.wrap_serializer_function_ser_schema(
                kls.__pydantic_serialization__
            ),
            ref=f"{kls.__name__}:{id(kls)}"
        )

    def __python_value_restore__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            input_value: typing.Any,
            info: ValidationInfo,
    ) -> typing.Any:
        from typenum.pydantic.core import TypEnumPydantic

        if isinstance(input_value, TypEnumPydantic):
            return input_value

        if isinstance(input_value, str):
            input_value = {input_value: None}

        for attr, value in input_value.items():
            attr = kls.__names_deserialization__.get(attr, attr)
            return getattr(kls, attr).__variant_constructor__(value, info)

    def __pydantic_serialization__(
            self,
            kls: type["TypEnumPydantic[TypEnumContent]"],
            model: typing.Any,
            serializer: SerializerFunctionWrapHandler,
    ) -> typing.Any:
        attr = model.__variant_name__
        attr = kls.__names_serialization__.get(attr, attr)

        if model.__content_type__ is NoValue:
            return attr
        elif isinstance(model.value, kls):
            content = kls.__pydantic_serialization__(model.value, serializer)
        else:
            content = serializer(model.value)

        return {attr: content}
